#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bio.h>
#include "der.h"

struct {
	int num;
	char *str;
} Tags[] = {
	{ BOOLEAN,				"BOOLEAN" },
	{ INTEGER,				"INTEGER" },
	{ BIT_STRING,			"BIT STRING" },
	{ OCTET_STRING,			"OCTET STRING" },
	{ NULLTAG,				"NULLTAG" },
	{ OBJECT_ID,			"OID" },
	{ ObjectDescriptor,		"OBJECT_DES" },
	{ EXTERNAL,				"EXTERNAL" },
	{ REAL,					"REAL" },
	{ ENUMERATED,			"ENUMERATED" },
	{ EMBEDDED_PDV,			"EMBEDDED PDV" },
	{ SEQUENCE,				"SEQUENCE" },
	{ SETOF,				"SETOF" },
	{ UTF8String,			"UTF8String" },
	{ NumericString,		"NumericString" },
	{ PrintableString,		"PrintableString" },
	{ TeletexString,		"TeletexString" },
	{ VideotexString,		"VideotexString" },
	{ IA5String,			"IA5String" },
	{ UTCTime,				"UTCTime" },
	{ GeneralizedTime,		"GeneralizedTime" },
	{ GraphicString,		"GraphicString" },
	{ VisibleString,		"VisibleString" },
	{ GeneralString,		"GeneralString" },
	{ UniversalString,		"UniversalString" },
	{ BMPString,			"BMPString" },
};

static char*
tagname(Tag tag)
{
	int i;
	char *ret;
	static char buf[32];

	ret = buf;
	switch(tag.class){
	case Universal:
		snprint(buf, sizeof(buf), "Universal_%d", tag.num);
		for(i = 0; i < nelem(Tags); i++)
			if(Tags[i].num == tag.num)
				ret = Tags[i].str;
		break;
	case Application:
		snprint(buf, sizeof(buf), "Application_%d", tag.num);
		break;
	case Context:
		snprint(buf, sizeof(buf), "[%d]", tag.num);
		break;
	case Private:
		snprint(buf, sizeof(buf), "Private_%d", tag.num);
		break;
	}
	return ret;
}

static void
hexdump(Biobuf *bp, uint indent, uchar *buf, int len)
{
	int i, j, base;

	base = 0;
	for(j = 0; j < len; j +=  i){
		Bprint(bp, "%*.s	%04x\t", indent, "", base +j);

		for(i = 0; i < 16 && j + i < len; i++)
			Bprint(bp, "%02x ", buf[j+i]);

		for(; i < 16; i++)
			Bprint(bp, "   ");

		Bprint(bp, "\t");

		for(i = 0; i < 16 && j + i < len; i++)
			if(isprint(buf[j+i]) && !isspace(buf[j+i]))
				Bprint(bp, "%c", buf[j+i]);
			else
				Bprint(bp, ".");

		Bprint(bp, "\n");
	}
	Bprint(bp, "\n");
}

/* could be confused by Unicode but it just tweeks the formatting so not a big issue */
static int
printable(uchar *buf, int len)
{
	int i;

	for(i = 0; i < len; i++)
		if(! isprint(buf[i]))
			return 0;
	return 1;
}

void
do_dump(Biobuf *bp, Elem e, int indent)
{
	Value v;
	Elist *el;
	int i, multi;

	Bprint(bp, "%*.s%s {", indent, "", tagname(e.tag));

	v = e.val;
	multi = 0;

	switch(v.tag){
	case VBool:
		Bprint(bp, " Bool %d", v.u.boolval);
		break;
	case VInt:
		Bprint(bp, " Int %d", v.u.intval);
		break;
	case VOctets:
		Bprint(bp, " Octets[%d]", v.u.octetsval->len);

		if(v.u.octetsval->len == 0)
			Bprint(bp, "''");
		else
		if(printable(v.u.bigintval->data, v.u.bigintval->len))
			Bprint(bp, "	'%.*s'", v.u.bigintval->len, (char *)v.u.bigintval->data);
		else{
			Bprint(bp, "\n");
			multi = 1;
			hexdump(bp, indent, v.u.bigintval->data, v.u.bigintval->len);
		}
		break;
	case VBigInt:
		Bprint(bp, " BigInt[%d] %.2x%.2x...", v.u.bigintval->len, v.u.bigintval->data[0], v.u.bigintval->data[1]);
		break;
	case VReal:
		Bprint(bp, " Real...");
		break;
	case VOther:
		Bprint(bp, " Other...");
		break;
	case VBitString:
		Bprint(bp, " BitString[%d]...", v.u.bitstringval->len * 8 - v.u.bitstringval->unusedbits);
		break;
	case VNull:
		Bprint(bp, " Null");
		break;
	case VEOC:
		Bprint(bp, " EOC...");
		break;
	case VObjId:
		Bprint(bp, " ObjId %d", v.u.objidval->data[0]);
		for(i = 1; i < v.u.objidval->len; i++)
			Bprint(bp, ".%d", v.u.objidval->data[i]);
		break;
	case VString:
		Bprint(bp, " String \"%s\"", v.u.stringval);
		break;
	case VSeq:
		Bprint(bp, "\n");
		for(el = v.u.seqval; el != nil; el = el->tl){
			multi = 1;
			do_dump(bp, el->hd, indent+4);
		}
		break;
	case VSet:
		Bprint(bp, "\n");
		for(el = v.u.setval; el != nil; el = el->tl){
			multi = 1;
			do_dump(bp, el->hd, indent+4);
		}
		break;
	}

	if(multi)
		Bprint(bp, "%*.s}\n", indent, "");
	else
		Bprint(bp, " }\n");

}

int
der_dump(uchar *der, int len)
{
	Elem e;
	int err;
	Biobuf out;

	if((err = der_decode(der, len, &e)) != ASN_OK){
		fprint(2, "der_dump: decode failed %d\n", err);
		return -1;
	}
	Binit(&out, 1, OWRITE);
	do_dump(&out, e, 0);
	Bterm(&out);
	freevalfields(&e.val);
	return 0;
}

